@@ -2,15 +2,26 @@ from __future__ import annotations
 
 import os
 import os.path
-from typing import TYPE_CHECKING, Any, Dict, List, Optional
+from typing import TYPE_CHECKING, Any, Dict, List, Optional, cast
 
 import boto3
+from typing_extensions import TypedDict
+
+from aec.util.ec2 import describe_running_instances_names
+from aec.util.errors import HandledError, NoInstancesError
+from aec.util.threads import executor
 
 if TYPE_CHECKING:
-    from mypy_boto3_ec2.type_defs import BlockDeviceMappingTypeDef, FilterTypeDef
+    from mypy_boto3_ec2.type_defs import (
+        BlockDeviceMappingTypeDef,
+        FilterTypeDef,
+        TagTypeDef,
+        InstanceStatusSummaryTypeDef,
+    )
+    from mypy_boto3_ec2.literals import InstanceTypeType
 
 import aec.command.ami as ami_cmd
-import aec.util.tags as util_tags
+import aec.util.ec2 as util_tags
 from aec.util.config import Config
 from aec.util.ec2_types import RunArgs
 
@@ -19,6 +30,15 @@ def is_ebs_optimizable(instance_type: str) -> bool:
     return not instance_type.startswith("t2")
 
 
+class Instance(TypedDict, total=False):
+    InstanceId: str
+    State: str
+    Name: Optional[str]
+    Type: str
+    DnsName: str
+    SubnetId: str
+
+
 def launch(
     config: Config,
     name: str,
@@ -29,13 +49,13 @@ def launch(
     instance_type: Optional[str] = None,
     key_name: Optional[str] = None,
     userdata: Optional[str] = None,
-) -> List[Dict[str, Any]]:
+) -> List[Instance]:
     """Launch a tagged EC2 instance with an EBS volume."""
 
     if not (template or ami):
         raise ValueError("Please specify either an ami or a launch template")
 
-    if not template and ami and not instance_type:
+    if not template and not instance_type:
         # if no instance type is provided set one
         instance_type = "t3.small"
 
@@ -82,12 +102,14 @@ def launch(
     key = key_name or config.get("key_name", None)
     if key:
         runargs["KeyName"] = key
+    elif not template:
+        raise HandledError("Please provide a key name")
 
     if instance_type:
-        runargs["InstanceType"] = instance_type
+        runargs["InstanceType"] = cast("InstanceTypeType", instance_type)
         runargs["EbsOptimized"] = is_ebs_optimizable(instance_type)
 
-    tags = [{"Key": "Name", "Value": name}]
+    tags: List[TagTypeDef] = [{"Key": "Name", "Value": name}]
     additional_tags = config.get("additional_tags", {})
     if additional_tags:
         tags.extend([{"Key": k, "Value": v} for k, v in additional_tags.items()])
@@ -151,38 +173,52 @@ def describe(
     include_terminated: bool = False,
     show_running_only: bool = False,
     sort_by: str = "State,Name",
-) -> List[Dict[str, Any]]:
+    columns: Optional[str] = None,
+) -> List[Instance]:
     """List EC2 instances in the region."""
 
     ec2_client = boto3.client("ec2", region_name=config.get("region", None))
 
-    response = ec2_client.describe_instances(Filters=filters(name, name_match))
+    filters = name_filters(name, name_match)
+    if show_running_only:
+        filters.append({"Name": "instance-state-name", "Values": ["pending", "running"]})
+
+    response = ec2_client.describe_instances(Filters=filters)
 
     # print(response["Reservations"][0]["Instances"][0])
 
-    instances: List[Dict[str, Any]] = [
-        {
-            "State": i["State"]["Name"],
-            "Name": util_tags.get_value(i, "Name"),
-            "Type": i["InstanceType"],
-            "DnsName": i["PublicDnsName"] if i.get("PublicDnsName", None) != "" else i["PrivateDnsName"],
-            "LaunchTime": i["LaunchTime"],
-            "ImageId": i["ImageId"],
-            "InstanceId": i["InstanceId"],
-        }
-        for r in response["Reservations"]
-        for i in r["Instances"]
-        if (include_terminated or i["State"]["Name"] != "terminated")
-        and (not show_running_only or i["State"]["Name"] in ["pending", "running"])
-    ]
+    cols = columns.split(",") if columns else ["InstanceId", "State", "Name", "Type", "DnsName"]
+
+    # don't sort by cols we aren't showing
+    sort_cols = [sc for sc in sort_by.split(",") if sc in cols]
+
+    instances: List[Instance] = []
+    for r in response["Reservations"]:
+        for i in r["Instances"]:
+            if include_terminated or i["State"]["Name"] != "terminated":
+                desc: Instance = {}
+
+                for col in cols:
+                    if col == "State":
+                        desc[col] = i["State"]["Name"]
+                    elif col == "Name":
+                        desc[col] = util_tags.get_value(i, "Name")
+                    elif col == "Type":
+                        desc[col] = i["InstanceType"]
+                    elif col == "DnsName":
+                        desc[col] = i["PublicDnsName"] if i.get("PublicDnsName", None) != "" else i["PrivateDnsName"]
+                    else:
+                        desc[col] = i[col]
+
+                instances.append(desc)
 
     return sorted(
         instances,
-        key=lambda i: "".join(str(i[field]) for field in sort_by.split(",")),
+        key=lambda i: "".join(str(i[field]) for field in sort_cols),
     )
 
 
-def tags(
+def describe_tags(
     config: Config,
     name: Optional[str] = None,
     name_match: Optional[str] = None,
@@ -196,6 +232,32 @@ def tags(
     return instance_tags(config, name, name_match, keys)
 
 
+def tag(
+    config: Config,
+    tags: List[str],
+    name: Optional[str] = None,
+    name_match: Optional[str] = None,
+) -> List[Dict[str, Any]]:
+    """Tag EC2 instance(s)."""
+    ec2_client = boto3.client("ec2", region_name=config.get("region", None))
+
+    tagdefs: List[TagTypeDef] = []
+    for t in tags:
+        parts = t.split("=")
+        tagdefs.append({"Key": parts[0], "Value": parts[1]})
+
+    instances = describe(config, name, name_match)
+
+    ids = [i["InstanceId"] for i in instances]
+
+    if not ids:
+        raise NoInstancesError(name=name, name_match=name_match)
+
+    ec2_client.create_tags(Resources=ids, Tags=tagdefs)
+
+    return describe_tags(config, name, name_match, keys=[d["Key"] for d in tagdefs])
+
+
 def instance_tags(
     config: Config, name: Optional[str] = None, name_match: Optional[str] = None, keys: List[str] = []
 ) -> List[Dict[str, Any]]:
@@ -203,7 +265,7 @@ def instance_tags(
 
     ec2_client = boto3.client("ec2", region_name=config.get("region", None))
 
-    response = ec2_client.describe_instances(Filters=filters(name, name_match))
+    response = ec2_client.describe_instances(Filters=name_filters(name, name_match))
 
     instances: List[Dict[str, Any]] = []
     for r in response["Reservations"]:
@@ -228,7 +290,7 @@ def volume_tags(
 
     ec2_client = boto3.client("ec2", region_name=config.get("region", None))
 
-    response = ec2_client.describe_volumes(Filters=filters(name, name_match))
+    response = ec2_client.describe_volumes(Filters=name_filters(name, name_match))
 
     volumes: List[Dict[str, Any]] = []
     for v in response["Volumes"]:
@@ -244,7 +306,7 @@ def volume_tags(
     return sorted(volumes, key=lambda i: str(i["Name"]))
 
 
-def start(config: Config, name: str) -> List[Dict[str, Any]]:
+def start(config: Config, name: str) -> List[Instance]:
     """Start EC2 instance."""
 
     ec2_client = boto3.client("ec2", region_name=config.get("region", None))
@@ -254,7 +316,7 @@ def start(config: Config, name: str) -> List[Dict[str, Any]]:
     instances = describe(config, name)
 
     if not instances:
-        raise Exception(f"No instances with {pretty_name_or_id(name)}")
+        raise NoInstancesError(name=name)
 
     instance_ids = [instance["InstanceId"] for instance in instances]
     ec2_client.start_instances(InstanceIds=instance_ids)
@@ -273,17 +335,13 @@ def stop(config: Config, name: str) -> List[Dict[str, Any]]:
     instances = describe(config, name)
 
     if not instances:
-        raise Exception(f"No instances with {pretty_name_or_id(name)}")
+        raise NoInstancesError(name=name)
 
     response = ec2_client.stop_instances(InstanceIds=[instance["InstanceId"] for instance in instances])
 
     return [{"State": i["CurrentState"]["Name"], "InstanceId": i["InstanceId"]} for i in response["StoppingInstances"]]
 
 
-def pretty_name_or_id(name: str) -> str:
-    return f"instance id {name}" if name.startswith("i-") else f"name {name}"
-
-
 def terminate(config: Config, name: str) -> List[Dict[str, Any]]:
     """Terminate EC2 instance."""
 
@@ -292,7 +350,7 @@ def terminate(config: Config, name: str) -> List[Dict[str, Any]]:
     instances = describe(config, name)
 
     if not instances:
-        raise Exception(f"No instances with {pretty_name_or_id(name)}")
+        raise NoInstancesError(name=name)
 
     response = ec2_client.terminate_instances(InstanceIds=[instance["InstanceId"] for instance in instances])
 
@@ -301,7 +359,7 @@ def terminate(config: Config, name: str) -> List[Dict[str, Any]]:
     ]
 
 
-def modify(config: Config, name: str, type: str) -> List[Dict[str, Any]]:
+def modify(config: Config, name: str, type: str) -> List[Instance]:
     """Change an instance's type."""
 
     ec2_client = boto3.client("ec2", region_name=config.get("region", None))
@@ -309,7 +367,7 @@ def modify(config: Config, name: str, type: str) -> List[Dict[str, Any]]:
     instances = describe(config, name)
 
     if not instances:
-        raise Exception(f"No instances with {pretty_name_or_id(name)}")
+        raise NoInstancesError(name=name)
 
     instance_id = instances[0]["InstanceId"]
     ec2_client.modify_instance_attribute(InstanceId=instance_id, InstanceType={"Value": type})
@@ -340,7 +398,7 @@ def logs(config: Config, name: str) -> str:
     instances = describe(config, name)
 
     if not instances:
-        raise Exception(f"No instances with {pretty_name_or_id(name)}")
+        raise NoInstancesError(name=name)
 
     instance_id = instances[0]["InstanceId"]
     response = ec2_client.get_console_output(InstanceId=instance_id)
@@ -361,7 +419,52 @@ def templates(config: Config) -> List[Dict[str, Any]]:
     ]
 
 
-def filters(name: Optional[str] = None, name_match: Optional[str] = None) -> List[FilterTypeDef]:
+def status(config: Config) -> List[Dict[str, Any]]:
+    """Describe instances status checks."""
+    ec2_client = boto3.client("ec2", region_name=config.get("region", None))
+
+    kwargs = {"MaxResults": 50}
+
+    response_fut = executor.submit(ec2_client.describe_instance_status, **kwargs)
+    instances = executor.submit(describe_running_instances_names, config).result()
+    response = response_fut.result()
+
+    statuses = []
+    while True:
+        statuses.extend(
+            [
+                {
+                    "InstanceId": i["InstanceId"],
+                    "State": i["InstanceState"]["Name"],
+                    "Name": instances.get(i["InstanceId"], None),
+                    "System status check": status_text(i["SystemStatus"]),
+                    "Instance status check": status_text(i["InstanceStatus"]),
+                }
+                for i in response["InstanceStatuses"]
+            ]
+        )
+
+        next_token = response.get("NextToken", None)
+        if next_token:
+            kwargs = {"NextToken": next_token}
+            response = ec2_client.describe_instance_status(**kwargs)
+        else:
+            break
+
+    return sorted(
+        statuses,
+        key=lambda i: "".join(str(i[field]) for field in ["State", "Name"]),
+    )
+
+
+def status_text(summary: InstanceStatusSummaryTypeDef, key: str = "reachability") -> str:
+    status = [d for d in summary["Details"] if d["Name"] == key][0]
+    return f"{status['Name']} {status['Status']}" + (
+        f" since {status['ImpairedSince']}" if status.get("ImpairedSince", None) else ""
+    )
+
+
+def name_filters(name: Optional[str] = None, name_match: Optional[str] = None) -> List[FilterTypeDef]:
     if name and name.startswith("i-"):
         return [{"Name": "instance-id", "Values": [name]}]
     elif name:
