@@ -1,7 +1,10 @@
 import argparse
 import sys
+import traceback
 from typing import List
 
+from botocore.exceptions import ClientError, NoCredentialsError
+
 import aec.command.ami as ami
 import aec.command.compute_optimizer as compute_optimizer
 import aec.command.ec2 as ec2
@@ -11,16 +14,26 @@ import aec.util.config as config
 import aec.util.configure as configure
 import aec.util.display as display
 from aec.util.cli import Arg, Cmd
+from aec.util.errors import HandledError
 
 config_arg = Arg("--config", help="Section of the config file to use")
 
 
 def ami_arg_checker(s: str) -> str:
     if not (s.startswith("ami-") or s in ami.ami_keywords.keys()):
-        raise argparse.ArgumentTypeError(f"must begin with 'ami-' or be one of {[k for k in ami.ami_keywords.keys()]}")
+        raise argparse.ArgumentTypeError(f"must begin with 'ami-' or be one of {list(ami.ami_keywords.keys())}")
+
     return s
 
 
+def tag_arg_checker(tag: str) -> str:
+    parts = tag.split("=")
+    if len(parts) != 2:
+        raise argparse.ArgumentTypeError(f"Invalid tag argument '{tag}'. Must be in key=value form.")
+
+    return tag
+
+
 # fmt: off
 
 configure_cli = [
@@ -40,6 +53,7 @@ ec2_cli = [
         Arg("-r", "--show-running-only", action='store_true', help="Show running or pending instances only"),
         Arg("-it", "--include-terminated", action='store_true', help="Include terminated instances"),
         Arg("-s", "--sort-by", type=str, help="Sort by one or more fields", default="State,Name"),
+        Arg("-c", "--columns", type=str, help="Customise the columns shown", default="InstanceId,State,Name,Type,DnsName,LaunchTime,ImageId"),
     ]),
     Cmd(ec2.launch, [
         config_arg,
@@ -69,12 +83,21 @@ ec2_cli = [
         config_arg,
         Arg("name", type=str, help="Name tag of instance or instance id")
     ]),
-    Cmd(ec2.tags, [
+    Cmd(ec2.tag, [
+        config_arg,
+        Arg("-t", "--tags", type=tag_arg_checker, nargs='+', metavar="TAG", help="Tags to create in key=value form", required = True),
+        Arg("name", type=str, nargs='?', help="Filter to instances with this Name tag or instance id."),
+        Arg("-q", type=str, dest='name_match', help="Filter to instances with a Name tag containing NAME_MATCH.")
+    ]),
+    Cmd(ec2.describe_tags, [
         config_arg,
         Arg("name", type=str, nargs='?', help="Filter to instances with this Name tag or instance id."),
         Arg("-q", type=str, dest='name_match', help="Filter to instances with a Name tag containing NAME_MATCH."),
         Arg("-v", "--volumes", action='store_true', help="Show volumes"),
-        Arg("-k", "--keys", type=str, nargs='*', metavar="KEY", help="Tags to display", default = [])
+        Arg("-k", "--keys", type=str, nargs='*', metavar="KEY", help="Tags to display", default = []),
+    ], name = "tags"),
+    Cmd(ec2.status, [
+        config_arg
     ]),
     Cmd(ec2.templates, [
         config_arg
@@ -168,8 +191,28 @@ def build_parser() -> argparse.ArgumentParser:
 
 
 def main(args: List[str] = sys.argv[1:]) -> None:
-    result, output_format = cli.dispatch(build_parser(), args)
-    display.pretty_print(result, output_format)
+    try:
+        result, output_format = cli.dispatch(build_parser(), args)
+        display.pretty_print(result, output_format)
+    except ClientError as e:
+        code = e.response["Error"]["Code"]
+        if code == "UnauthorizedOperation":
+            message = e.response["Error"]["Message"]
+            print(f"{code}: {message}\n\nAuthenticate with the appropriate AWS role before retrying.", file=sys.stderr)
+
+        elif code == "RequestExpired":
+            print(
+                f"{code}: AWS session token expired.\n\nRe-authenticate with the appropriate AWS role.", file=sys.stderr
+            )
+
+        else:
+            traceback.print_exc(file=sys.stderr)
+    except NoCredentialsError as e:
+        print(
+            f"NoCredentialsError: {e}.\n\nAuthenticate with the appropriate AWS role before retrying.", file=sys.stderr
+        )
+    except HandledError as e:
+        print(e, file=sys.stderr)
 
 
 if __name__ == "__main__":
